__all__ = [
    "utils",
    "Factory",
    "pipelines",
    "scripts",
    "workflows",
    "nodes",
    "pan_processing",
    "single_subject",
    "group_subjects",
    "version"
]

from . import version
__version__ = version.__version__